// Doc.cpp


#include "stdafx.h"
#include "SimuRoute.h"

#include "Doc.h"
#include "GraphView.h"
#include "DlgProperties.h"
#include "DlgSize.h"


#ifdef _DEBUG
#define new DEBUG_NEW
#endif


const CSize aoffDir[8] =
{
	CSize( 0,-1), CSize(+1, 0), CSize( 0,+1), CSize(-1, 0),   // T, R, B, L
	CSize(+1,-1), CSize(+1,+1), CSize(-1,+1), CSize(-1,-1),   // TR, BR, BL, TL
};
const CSize aoffDiagDir[4] =
{
	CSize(+1,0), CSize(+1,+1), CSize(0,+1), CSize(0,0),
};

const CRect arcBorder[4] =
{
	CRect(0,0, 1,0), CRect(1,0, 1,1), CRect(0,1, 1,1), CRect(0,0, 0,1),
};



IMPLEMENT_DYNCREATE(CDoc, CDocument)
BEGIN_MESSAGE_MAP(CDoc, CDocument)
	ON_COMMAND          (ID_EDIT_CLEAR,      OnEditClear)
	ON_UPDATE_COMMAND_UI(ID_EDIT_CLEAR,      OnUpdateNeedSel)
	ON_COMMAND          (ID_EDIT_SELECT_ALL, OnEditSelectAll)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SELECT_ALL, OnUpdateEditSelectAll)
	ON_COMMAND          (ID_EDIT_UNSELECT,   OnEditUnselect)
	ON_UPDATE_COMMAND_UI(ID_EDIT_UNSELECT,   OnUpdateNeedSel)
	ON_COMMAND_RANGE          (idToolFirst,  idToolLast,  OnGraphTool)
	ON_UPDATE_COMMAND_UI_RANGE(idToolFirst,  idToolLast,  OnUpdateGraphTool)
	ON_COMMAND(ID_GRAPH_SIZE, OnGraphSize)
END_MESSAGE_MAP()


// Constructor
CDoc::CDoc() : m_aCell(NULL)
{
}

// Destructor
CDoc::~CDoc()
{
	ASSERT(m_lstNode.IsEmpty());
	
	delete [] m_aCell;
}


// Delete document contents
void CDoc::DeleteContents()
{
	POSITION pos = m_lstNode.GetHeadPosition();
	while (pos)
		delete m_lstNode.GetNext(pos);
	m_lstNode.RemoveAll();
	
	delete [] m_aCell;
	m_aCell = NULL;
	m_size.SetSize(20,20);
	OnUpdateBoardSize();
	
	
	CDocument::DeleteContents();
}


// Serialization
void CDoc::Serialize(CArchive& ar)
{
	if (ar.IsStoring())
		// Saving
		ar << m_size;
	
	else{
		// Loading
		ar >> m_size;
		
		OnUpdateBoardSize();
	}
	
	
	m_lstNode.Serialize(ar);
	
	if (ar.IsLoading()) {
		// Loading
		
		POSITION pos = m_lstNode.GetHeadPosition();
		while (pos) {
			const POSITION posSauve = pos;
			if (!m_lstNode.GetNext(pos))
				m_lstNode.RemoveAt(posSauve);
		}
		
		UpdateCells(false);
	}
}


// Diagnostics
#ifdef _DEBUG
void CDoc::AssertValid() const
{
	CDocument::AssertValid();
	m_lstNode.AssertValid();
}

void CDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
	m_lstNode.Dump(dc);
}
#endif


// Indicate if there is an overlap problem
// If fIgnoreSel==true, we ignore overlap with selected nodes
bool CDoc::IsOverlapOK(const CNode& node, bool fIgnoreSel) const
{
	const CRect rcNode = node.GetCellsRect();
	
	// Screen limits
	if (rcNode.left<0 || rcNode.top<0 ||
	    rcNode.right>m_size.cx || rcNode.bottom>m_size.cy)
		return false;
	
	// Overlap
	POSITION pos = m_lstNode.GetHeadPosition();
	while (pos) {
		CNode *const pNode = m_lstNode.GetNext(pos);
		if (pNode!=&node && (!fIgnoreSel || !pNode->m_sel))
			VERIF((rcNode & pNode->GetCellsRect()).IsRectEmpty());
	}
	
	return true;
}


// Add a new node
void CDoc::AddNode(const CNode& node)
{
	if (!IsOverlapOK(node))
		return;
	
	ASSERT(node.m_type!=nodeInvalid);
	
	CNode *const pNode = new CNode;
	*pNode = node;
	m_lstNode.AddTail(pNode);
	
	UpdateCells();
}


// Delete selection
void CDoc::OnEditClear()
{
	POSITION pos = m_lstNode.GetHeadPosition();
	while (pos) {
		const POSITION posSave = pos;
		CNode *const pNode = m_lstNode.GetNext(pos);
		if (pNode->m_sel) {
			m_lstNode.RemoveAt(posSave);
			delete pNode;
		}
	}
	
	UpdateCells();
}


// An node must be selected
void CDoc::OnUpdateNeedSel(CCmdUI* pCmdUI)
{
	bool fSel = false;
	POSITION pos = m_lstNode.GetHeadPosition();
	while (pos) {
		if (m_lstNode.GetNext(pos)->m_sel) {
			fSel = true;
			break;
		}
	}
	
	pCmdUI->Enable(fSel);
}


// Current tool
void CDoc::OnGraphTool(UINT id)
{
	CGraphView::s_idTool = id;
	
	if (id>=idNodeFirst && id<=idNodeLast)
		SetAllSelect(selNo);
	
	UpdateAllViews(NULL, updateOnToolChange);
}

// Current tool - Update
void CDoc::OnUpdateGraphTool(CCmdUI* pCmdUI)
{
	pCmdUI->SetRadio(pCmdUI->m_nID==CGraphView::s_idTool);
}


// Board size modified
void CDoc::OnUpdateBoardSize()
{
	delete [] m_aCell;
	m_aCell = new CELL[m_size.cx * m_size.cy];
	
	UpdateAllViews(NULL, updateBoardSize);
	UpdateCells();
}


// Selection modified -> update properties dialog box
void CDoc::OnSelectionChanged()
{
	e_dlgProp.Update(this);
}


// Update the array "cell position -> node"
void CDoc::UpdateCells(bool fDocModified)
{
	ASSERT(m_aCell);
	
	ZeroMemory(m_aCell, m_size.cx*m_size.cy * sizeof(m_aCell[0]));
	
	POSITION pos = m_lstNode.GetHeadPosition();
	while (pos)
		m_lstNode.GetNext(pos)->UpdateCells(m_aCell, m_size);
	
	FindErrors();
	UpdateAllViews(NULL, updateRedrawAll);
	
	if (fDocModified)
		SetModifiedFlag();
}


// Find the errors in the current nodes
void CDoc::FindErrors()
{
	m_lstError.RemoveAll();
	
	for (int x=0; x<m_size.cx; x++) {
		for (int y=0; y<m_size.cy; y++) {
			const CELL &cell = GetCell(x,y);
			if (!cell.IsRoad())
				continue;
			
			for (int iOrient=0, iWall=1; iOrient<4; iOrient++, iWall*=2) {
				if (!(cell.iWall & iWall)) {
					// No wall here: there must be something in the next cell
					
					CPoint ptNext = CPoint(x,y) + aoffDir[iOrient];
					const CELL *const pCellNext = GetSafeCell(ptNext);
					if (pCellNext) {
						// Point in board: verify the cell
						// There must be a node, and no wall in the opposite direction
						
						if (pCellNext->IsRoad() &&
						   !(pCellNext->iWall & (1<<RoundOrient(iOrient+2))))
							// No problem
							continue;
					}
					
					// Problem here!
					// Compute coordinates of the border
					CRect rc(arcBorder[iOrient] + CPoint(x,y));
					m_lstError.AddTail(rc);
				}
			}
		}
	}
}


// Set the "m_sel" member for all nodes
void CDoc::SetAllSelect(NODESEL sel)
{
	POSITION pos = m_lstNode.GetHeadPosition();
	while (pos)
		m_lstNode.GetNext(pos)->m_sel = sel;
	
	UpdateAllViews(NULL, updateRedrawAll);
	OnSelectionChanged();
}


// Select all
void CDoc::OnEditSelectAll()
{
	SetAllSelect(selYes);
}

// Select all - Update
void CDoc::OnUpdateEditSelectAll(CCmdUI* pCmdUI)
{
	pCmdUI->Enable(m_lstNode.GetCount() > 0);
}


// Unselect
void CDoc::OnEditUnselect()
{
	SetAllSelect(selNo);
}


// Grid size
void CDoc::OnGraphSize()
{
	// Compute the minimal size
	CSize sizeMin(0,0);
	POSITION pos = m_lstNode.GetHeadPosition();
	while (pos) {
		CRect rc = m_lstNode.GetNext(pos)->GetCellsRect(true);
		sizeMin.cx = max(sizeMin.cx, rc.right);
		sizeMin.cy = max(sizeMin.cy, rc.bottom);
	}
	
	CDlgSize dlgSize;
	dlgSize.m_size    = m_size;
	dlgSize.m_sizeMin = sizeMin;
	if (dlgSize.DoModal()==IDOK) {
		m_size = dlgSize.m_size;
		OnUpdateBoardSize();
	}
}
