// GraphView.h


#pragma once

class CDoc;
struct CELL;
struct UNIT;


#include "Node.h"
#include "Resource.h"


const int cBorder      =  2;
const int cBorderInt   =  1;
const int cBorderError =  2;

const int cArrow       = 10;
const int cCar         = cArrow;


struct STATISTICS
{
	int   time;             // Simulation time
	int   nbCarCreated;
	int   nbCarDestroyed;
	int   nbCarRefused;     // Car creations refused
	int   timeCarLive;      // Total living time for destroyed cars
	int   nbSnail;
	int   nbSnailCreated;
	
	// In the last frames only
	float avgCarCreated;
	float avgCarRefused;
	float avgCarLive;
	float avgRefused;
};


enum
{
	updateOther,
	updateBoardSize,
	updateRedrawNode,
	updateRedrawAll,
	updateOnToolChange,
};


const UINT idToolFirst = ID_GRAPH_TOOL_HAND;
const UINT idToolLast  = ID_GRAPH_TOOL_NODE_LINE;

const UINT idNodeFirst = ID_GRAPH_TOOL_NODE_END;
const UINT idNodeLast  = ID_GRAPH_TOOL_NODE_LINE;


#include "Simul.h"


class CGraphView : public CScrollView
{
protected:  // Creation
	
	CGraphView();
	DECLARE_DYNCREATE(CGraphView)
	
	
public:  // Functions
	
	CDoc* GetDocument() const;
	
	virtual ~CGraphView();
	
	
public:  // Functions
	
	void SetCellSize(int cCell);
	void OnUpdateBoardSize();
	
	inline CPoint GetConvert(const POINT& pt) const;
	inline void Convert(POINT& pt) const;
	inline void Convert(CRect&  rc) const;
	
	inline void ToCell(POINT& pt) const
	{
		pt.x /= m_cCell;
		pt.y /= m_cCell;
	}
	
	inline CPoint GetToCell(const POINT& pt) const
	{
		return CPoint(pt.x/m_cCell, pt.y/m_cCell);
	}
	
	
	CRect GetViewRect() const
	{
		return CRect(CPoint(0,0), m_size);
	}
	
	
	void ComputeBorderRect(CRect arc[], int cBorder);
	void ComputeCornerRect(CRect arc[], int cBorder);
	
	void DrawCell (CDC& dc, int x, int y, const CELL& cell, bool fSimul) const;
	void DrawCells(CDC& dc, CELL aCell[], CRect rc, bool fSimul) const;
	
	
public:  // Data
	
	int   m_cCell;         // Cells dimensions in pixels
	int   m_cUnit;         // Units dimensions in pixels
	bool  m_fSimul;        // Simulation mode?
	int   m_fGrid;         // Show grid?
	bool  m_fViewErrors;   // View errors?
	CSize m_size;          // Usable size
	
	
	CELL *m_aCellShadow;
	
	CRect m_arcCellBorder[4], m_arcCellBorderInt[4], m_arcCellBorderLimit[4];
	CRect m_arcCellCorner[4], m_arcCellCornerInt[4];
	
	
	// Dimension 1 -> OBJECTSEL
	// Dimension 2 -> the enum
	enum
	{
		brushRoad,
		brushBorder,
		brushBorderInt,
		brushHatched,
		
		brushCount
	};
	static const COLORREF   s_acl[selCount][brushCount-1];
	static       CBrush     s_abrHatched[selCount];
	static       CBrush     s_brError;
	static       CImageList s_imlArrow;
	
	
	// Offscreen bitmap and DC
	class COffscreen
	{
	public:  // Functions
		
		~COffscreen()
		{
			Delete();
		}
		
		void Create(CDC* pDC, CSize size);
		void Delete();
		
		
	public:  // Data
		
		CDC     m_dc;
		CBitmap m_bmp;
	};
	
	COffscreen m_draw;
	COffscreen m_road;
	
	
	void DrawRoads(CDC& dc, bool fSimul) const;
	
	
	// Mouse, tools
	CPoint m_ptMouseOld, m_ptStart;
	CNode  m_nodeCreate;
	bool   m_fMove;   // If selection tool: move or select
	CSize  m_sizeMoveOld;
	
	
public:  // Simulation
	
	void SimulUpdate(bool fOnlyDraw);
	void SimulTick();
	void SimulUpdateTimer(int iSimulSpeed);
	
	int m_iSimulSpeed;
	int m_iSimulSkip;
	
	bool  m_fLogging;
	CFile m_fileLogging;
	
	
	bool             m_fSimulRun;
	CSize            m_sizeUnit;
	UNIT            *m_aUnit;
	int              m_nbUnitValid;
	CAR            **m_apCar, **m_apCarNew;
	
	UNIT           **m_apUnitValid;   // Valid units (points to m_aUnit elements)
	
	CAR             *m_aCar;
	int              m_nbCar;
	int              m_nbWantSnail;   // Number of snails to create
	
	CList<GENERATOR> m_lstGen;
	int              m_idCar;
	STATISTICS       m_stats;
	
	CList<CLight>    m_lstLight;
	CLight& AddLight(const NODEDATA& nd, POINT ptDraw, int dirDraw, bool fGreen);
	
	
	// Access
	UNIT& GetUnit(int x, int y)
	{
		ASSERT(m_aUnit);
		return m_aUnit[x + y*m_sizeUnit.cx];
	}
	const UNIT& GetUnit(int x, int y) const
	{
		ASSERT(m_aUnit);
		return m_aUnit[x + y*m_sizeUnit.cx];
	}
	inline UNIT& GetUnit(const CPoint& pt)
	{
		return GetUnit(pt.x, pt.y);
	}
	inline const UNIT& GetUnit(const CPoint& pt) const
	{
		return GetUnit(pt.x, pt.y);
	}
	
	
	inline CAR* GetCar(CAR* apCar[], const CPoint& pt)
	{
		return apCar[pt.x + pt.y*m_sizeUnit.cx];
	}
	inline void SetCarPointer(CAR* apCar[], const CPoint& pt, CAR* pCar)
	{
		apCar[pt.x + pt.y*m_sizeUnit.cx] = pCar;
	}
	void SetCar(CAR* apCar[], const CPoint& pt, CAR* pCar);
	
	
	void ToUnit(CPoint& pt) const
	{
		pt.x *= 2;
		pt.y *= 2;
	}
	CPoint GetToUnit(const CPoint pt) const
	{
		return CPoint(pt.x*2, pt.y*2);
	}
	
	CPoint GetUnitView(int x, int y) const
	{
		return CPoint(x*m_cUnit, y*m_cUnit);
	}
	CPoint GetUnitView(const CPoint& pt) const
	{
		return GetUnitView(pt.x, pt.y);
	}
	
	
	void ToUnit(CPoint& pt, int iDiagDir) const;
	CPoint GetToUnit(const CPoint pt, int iDiagDir) const;
	
	void SetUnit(CPoint pt, int maxSpeed, CLight* pLight,
		int dir0=-1, int proba0=0,
		int dir1=-1, int proba1=0,
		int dir2=-1, int proba2=0);
	
	void SetFourUnits(CPoint ptCell, int iDir, int maxSpeed, CLight* pLight);
	
	
	bool IsLogFileValid() const
	{
		return (HANDLE)m_fileLogging!=CFile::hFileNull;
	}
	
	void LogFileClose();
	void LogFileOpen(LPCTSTR pszFile);
	
	
	
public:  // Messages
	
	static UINT s_idTool;
	
	virtual BOOL PreCreateWindow(CREATESTRUCT& cs);
	virtual void OnInitialUpdate();
	virtual void OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint);
	afx_msg BOOL OnEraseBkgnd(CDC* pDC);
	virtual void OnDraw(CDC* pDC);
	virtual void OnActivateView(BOOL bActivate, CView* pActivateView, CView* pDeactiveView);
	afx_msg void OnViewGrid();
	afx_msg void OnUpdateViewGrid(CCmdUI* pCmdUI);
	afx_msg void OnLButtonDown(UINT nFlags, CPoint pt);
	afx_msg void OnRButtonDown(UINT nFlags, CPoint pt);
	afx_msg void OnMouseMove  (UINT nFlags, CPoint pt);
	afx_msg void OnLButtonUp  (UINT nFlags, CPoint pt);
	afx_msg BOOL OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT uMsg);
	afx_msg void OnGraphTurn(UINT id);
	afx_msg void OnGraphTurnUpdate(CCmdUI* pCmdUI);
	afx_msg void OnViewZoom(UINT id);
	afx_msg void OnUpdateViewZoom(CCmdUI* pCmdUI);
	afx_msg void OnViewErrors();
	afx_msg void OnUpdateViewErrors(CCmdUI* pCmdUI);
	afx_msg void OnSimulActivate();
	afx_msg void OnUpdateSimulActivate(CCmdUI* pCmdUI);
	afx_msg void OnSimulPause();
	afx_msg void OnUpdateSimulPause(CCmdUI* pCmdUI);
	afx_msg void OnTimer(UINT nIDEvent);
	afx_msg void OnUpdateIndicatorCoord(CCmdUI* pCmdUI);
	afx_msg void OnUpdateIndicatorUnit(CCmdUI* pCmdUI);
	afx_msg void OnSimulAddSnail();
	afx_msg void OnUpdateSimulAddSnail(CCmdUI* pCmdUI);
	afx_msg void OnSimulLogging();
	afx_msg void OnUpdateSimulLogging(CCmdUI* pCmdUI);
	afx_msg void OnSimulLogFile();
	afx_msg void OnPerso(UINT id);
	DECLARE_MESSAGE_MAP()
};


#ifndef _DEBUG
inline CDoc* CGraphView::GetDocument() const
	{ return (CDoc*)m_pDocument; }
#endif

// Convert coordinates from cell to pixels
inline CPoint CGraphView::GetConvert(const POINT& pt) const
{
	return CPoint(pt.x * m_cCell, pt.y * m_cCell);
}

// Idem for a reference
inline void CGraphView::Convert(POINT& pt) const
{
	pt.x *= m_cCell;
	pt.y *= m_cCell;
}

// Idem for a rectangle
void CGraphView::Convert(CRect& rc) const
{
	Convert(rc.TopLeft());
	Convert(rc.BottomRight());
}
